import pytest
import tempfile
import os
import numpy as np
from biosaic._rna import RNA

# DNA Tests
class TestDNAInitialization:
  def test_init_continuous(self):
    rna = RNA(kmer=3, continuous=True)
    assert rna.kmer == 3
    assert rna.continuous == True
    assert rna.vocab_size == 5**3
    assert rna.has_special_tokens == False

  def test_init_non_continuous(self):
    rna = RNA(kmer=3, continuous=False)
    assert rna.kmer == 3
    assert rna.continuous == False
    expected_size = 5 + 25 + 125 + len(rna.special_tokens)
    assert rna.vocab_size == expected_size

  def test_init_with_special_tokens(self):
    special_tokens = ['<START>', '<END>']
    rna = RNA(kmer=2, continuous=False, special_tokens=special_tokens)
    assert rna.has_special_tokens == True
    assert rna.special_tokens == special_tokens
    expected_size = (5 + 25) + 2
    assert rna.vocab_size == expected_size

  def test_special_tokens_with_continuous_raises_error(self):
    with pytest.raises(ValueError):
      RNA(kmer=3, continuous=True, special_tokens=['<START>'])

class TestDNATokenization:
  def test_tokenize_continuous(self):
    rna = RNA(kmer=3, continuous=True)
    tokens = rna.tokenize("AUGCAUGC")
    assert tokens == ['AUG', 'UGC', 'GCA', 'CAU', 'AUG', 'UGC']

  def test_tokenize_non_continuous(self):
    rna = RNA(kmer=3, continuous=False)
    tokens = rna.tokenize("AUGCAUGC")
    assert tokens == ['AUG', 'CAU', 'GC']

  def test_tokenize_with_special_tokens(self):
    rna = RNA(kmer=2, continuous=False, special_tokens=['<S>', '</S>'])
    tokens = rna.tokenize("<S>AUGC</S>")
    assert tokens == ['<S>', 'AU', 'GC', '</S>']

  def test_tokenize_invalid_character(self):
    rna = RNA(kmer=2)
    with pytest.raises(ValueError):
      rna.tokenize("AUGCX")

class TestDNAVocabulary:
  def test_build_vocab_continuous(self):
    rna = RNA(kmer=2, continuous=True)
    rna.build_vocab()
    assert len(rna.vocab) == 25
    assert 'AA' in rna.vocab
    assert 'UU' in rna.vocab
    assert '--' in rna.vocab

  def test_build_vocab_non_continuous(self):
    rna = RNA(kmer=2, continuous=False)
    rna.build_vocab()
    assert len(rna.vocab) == 35
    assert 'A' in rna.vocab
    assert 'AA' in rna.vocab

  def test_build_vocab_with_special_tokens(self):
    rna = RNA(kmer=2, continuous=False, special_tokens=['<S>', '</S>'])
    rna.build_vocab()
    assert len(rna.vocab) == 32  # 30 + 2
    assert '<S>' in rna.vocab
    assert '</S>' in rna.vocab

class TestDNAEncodeDecode:
  def test_encode_decode_roundtrip(self):
    rna = RNA(kmer=2, continuous=True)
    rna.build_vocab()
    sequence = "AUGCAUGC"
    encoded = rna.encode(sequence)
    decoded = rna.decode(encoded)
    assert decoded == sequence

  def test_encode_decode_with_special_tokens(self):
    rna = RNA(kmer=2, continuous=False, special_tokens=['<S>', '</S>'])
    rna.build_vocab()
    sequence = "<S>AUGC</S>"
    encoded = rna.encode(sequence)
    decoded = rna.decode(encoded)
    assert decoded == sequence

  def test_ids_to_chars_and_chars_to_ids(self):
    rna = RNA(kmer=2, continuous=True)
    rna.build_vocab()
    chars = ['AU', 'UG', 'GC']
    ids = rna.chars_to_ids(chars)
    converted_chars = rna.ids_to_chars(ids)
    assert chars == converted_chars

class TestDNADetokenization:
  def test_detokenize_continuous(self):
    rna = RNA(kmer=3, continuous=True)
    tokens = ['AUG', 'UGC', 'GCA']
    result = rna.detokenize(tokens)
    assert result == "AUGCA"
    
    assert rna.detokenize(['AUG']) == "AUG"
    assert rna.detokenize([]) == ""

  def test_detokenize_non_continuous(self):
    rna = RNA(kmer=2, continuous=False)
    tokens = ['AU', 'GC', 'UA']
    result = rna.detokenize(tokens)
    assert result == "AUGCUA"

class TestDNAVerification:
  def test_verify_continuous_valid(self):
    rna = RNA(kmer=3, continuous=True)
    rna.build_vocab()
    tokens = ['AUG', 'UGC', 'GCA']
    verified = rna.verify(tokens)
    assert verified[0]['match'] == True
    assert verified[1]['match'] == True

  def test_verify_continuous_invalid(self):
    rna = RNA(kmer=3, continuous=True)
    rna.build_vocab()
    tokens = ['AUG', 'CAU']
    verified = rna.verify(tokens)
    assert verified[0]['match'] == False

  def test_verify_with_special_tokens(self):
    rna = RNA(kmer=2, continuous=False, special_tokens=['<S>'])
    rna.build_vocab()
    tokens = ['<S>', 'AU', 'GC']
    verified = rna.verify(tokens)
    assert verified[0]['match'] == 'special_token'
    assert verified[1]['match'] == True

class TestDNAPersistence:
  @pytest.fixture
  def temp_dir(self):
    temp_dir = tempfile.mkdtemp()
    yield temp_dir
    for file in os.listdir(temp_dir):
      os.remove(os.path.join(temp_dir, file))
    os.rmdir(temp_dir)

  def test_save_and_load_json(self, temp_dir):
    rna = RNA(kmer=2, continuous=True)
    rna.build_vocab()
    save_path = os.path.join(temp_dir, "test_vocab")
    rna.save(save_path, as_json=True)
    
    dna2 = RNA(kmer=1)
    dna2.load(save_path + ".json")
    
    assert rna.kmer == dna2.kmer
    assert rna.continuous == dna2.continuous
    assert rna.vocab == dna2.vocab

  def test_save_and_load_pickle(self, temp_dir):
    rna = RNA(kmer=2, continuous=False, special_tokens=['<TEST>'])
    rna.build_vocab()
    save_path = os.path.join(temp_dir, "test_vocab")
    rna.save(save_path, as_json=False)
    
    dna2 = RNA(kmer=1)
    dna2.load(save_path + ".model")
    
    assert rna.kmer == dna2.kmer
    assert rna.continuous == dna2.continuous
    assert rna.special_tokens == dna2.special_tokens
    assert rna.vocab == dna2.vocab

class TestDNAUtilities:
  def test_one_hot_encode(self):
    rna = RNA(kmer=2, continuous=False)
    rna.build_vocab()
    one_hot = rna.one_hot_encode("AUGC")
    
    expected_shape = (2, len(rna.vocab))
    assert one_hot.shape == expected_shape
    for row in one_hot:
      assert np.sum(row) == 1

  def test_reverse_complement(self):
    rna = RNA(kmer=2)
    assert rna.reverse_complement("AUGC") == "GCAU"
    assert rna.reverse_complement("AU-GC") == "GC-AU"
    assert rna.reverse_complement("AUAU") == "AUAU"

  def test_pad_sequence(self):
    rna = RNA(kmer=2)
    assert rna.pad_sequence("AUG", 6) == "AUG---"
    assert rna.pad_sequence("AUGCAUGC", 5) == "AUGCA"
    assert rna.pad_sequence("AUG", 3) == "AUG"
    assert rna.pad_sequence("AU", 5, pad_char='N') == "AUNNN"

class TestDNAEdgeCases:
  def test_edge_cases(self):
    rna = RNA(kmer=1, continuous=True)
    rna.build_vocab()
    tokens = rna.tokenize("AUGC")
    assert tokens == ['A', 'U', 'G', 'C']
    
    rna = RNA(kmer=2, continuous=False)
    assert rna.tokenize("") == []
    assert rna.tokenize("A") == ['A']

if __name__ == '__main__':
  pytest.main([__file__, "-v"])


from ._dna import DNA
from ._rna import RNA
from ._main import Tokenizer
from ._protein import Protein
from .visual.cmap import CMap
from .visual.cgr import CGR

__version__ = "0.1.5"
__author__ = "Shivendra S"


import pytest
from unittest.mock import Mock, patch
from biosaic import Tokenizer

class TestTokenizerInitialization:
  def test_dna_tokenizer_init(self):
    with patch('biosaic._main.DNA') as mock_dna:
      mock_instance = Mock()
      mock_instance.has_special_tokens = False
      mock_dna.return_value = mock_instance
      
      tokenizer = Tokenizer("dna", 3, continuous=False)
      assert tokenizer.kmer == 3
      assert tokenizer.continuous == False
      assert tokenizer.encoding == "dna/base_3k"
      mock_dna.assert_called_once_with(kmer=3, continuous=False, special_tokens=None)

  def test_protein_tokenizer_init(self):
    with patch('biosaic._main.Protein') as mock_protein:
      mock_instance = Mock()
      mock_instance.has_special_tokens = True
      mock_protein.return_value = mock_instance
      
      tokenizer = Tokenizer("protein", 2, continuous=True, special_tokens=['<s>', '</s>'])
      assert tokenizer.kmer == 2
      assert tokenizer.continuous == True
      assert tokenizer.encoding == "protein/special_2k"
      mock_protein.assert_called_once_with(kmer=2, continuous=True, special_tokens=['<s>', '</s>'])

  def test_invalid_mode_raises_error(self):
    with pytest.raises(AssertionError, match="Unknown mode type"):
      Tokenizer("rna", 3)

  def test_dna_kmer_limit(self):
    with pytest.raises(AssertionError, match="KMer size supported only till 8"):
      Tokenizer("dna", 9)

  def test_protein_kmer_limit(self):
    with pytest.raises(AssertionError, match="KMer size supported only till 4"):
      Tokenizer("protein", 5)

  def test_encoding_path_construction(self):
    with patch('biosaic._main.DNA') as mock_dna:
      mock_instance = Mock()
      mock_instance.has_special_tokens = False
      mock_dna.return_value = mock_instance
      
      tokenizer = Tokenizer("dna", 4, continuous=True)
      expected_path = "https://raw.githubusercontent.com/delveopers/biosaic/main/vocab/dna/cont_4k.model"
      assert tokenizer.encoding_path == expected_path

class TestTokenizerMethods:
  @pytest.fixture
  def mock_tokenizer(self):
    with patch('biosaic._main.DNA') as mock_dna_class:
      mock_dna = Mock()
      mock_dna.has_special_tokens = False
      mock_dna.vocab_size = 64
      mock_dna.vocab = {'ATG': 0, 'TGC': 1, 'GCA': 2}
      mock_dna_class.return_value = mock_dna
      
      tokenizer = Tokenizer("dna", 3)
      return tokenizer, mock_dna

  def test_encode(self, mock_tokenizer):
    tokenizer, mock_dna = mock_tokenizer
    mock_dna.encode.return_value = [0, 1, 2]
    
    result = tokenizer.encode("ATGTGCGCA")
    assert result == [0, 1, 2]
    mock_dna.encode.assert_called_once_with("ATGTGCGCA")

  def test_decode(self, mock_tokenizer):
    tokenizer, mock_dna = mock_tokenizer
    mock_dna.decode.return_value = "ATGTGCGCA"
    
    result = tokenizer.decode([0, 1, 2])
    assert result == "ATGTGCGCA"
    mock_dna.decode.assert_called_once_with([0, 1, 2])

  def test_tokenize(self, mock_tokenizer):
    tokenizer, mock_dna = mock_tokenizer
    mock_dna.tokenize.return_value = ["ATG", "TGC", "GCA"]
    
    result = tokenizer.tokenize("ATGTGCGCA")
    assert result == ["ATG", "TGC", "GCA"]
    mock_dna.tokenize.assert_called_once_with("ATGTGCGCA")

  def test_detokenize(self, mock_tokenizer):
    tokenizer, mock_dna = mock_tokenizer
    mock_dna.detokenize.return_value = "ATGTGCGCA"
    
    result = tokenizer.detokenize(["ATG", "TGC", "GCA"])
    assert result == "ATGTGCGCA"
    mock_dna.detokenize.assert_called_once_with(["ATG", "TGC", "GCA"])

  def test_one_hot(self, mock_tokenizer):
    tokenizer, mock_dna = mock_tokenizer
    mock_dna.one_hot_encode.return_value = [[1, 0, 0, 0], [0, 1, 0, 0]]
    
    result = tokenizer.one_hot("AT")
    assert result == [[1, 0, 0, 0], [0, 1, 0, 0]]
    mock_dna.one_hot_encode.assert_called_once_with("AT")

  def test_reverse_complement(self, mock_tokenizer):
    tokenizer, mock_dna = mock_tokenizer
    mock_dna.reverse_complement.return_value = "CGCAT"
    
    result = tokenizer.reverse_complement("ATGCG")
    assert result == "CGCAT"
    mock_dna.reverse_complement.assert_called_once_with("ATGCG")

  def test_pad_sequence(self, mock_tokenizer):
    tokenizer, mock_dna = mock_tokenizer
    mock_dna.pad_sequence.return_value = "ATGC--"
    
    result = tokenizer.pad_sequence("ATGC", 6, "-")
    assert result == "ATGC--"
    mock_dna.pad_sequence.assert_called_once_with("ATGC", 6, "-")

  def test_vocab_size_property(self, mock_tokenizer):
    tokenizer, mock_dna = mock_tokenizer
    assert tokenizer.vocab_size == 64

  def test_vocab_property(self, mock_tokenizer):
    tokenizer, mock_dna = mock_tokenizer
    expected_vocab = {'ATG': 0, 'TGC': 1, 'GCA': 2}
    assert tokenizer.vocab == expected_vocab

class TestSpecialTokens:
  def test_default_special_tokens(self):
    with patch('biosaic._main.DNA') as mock_dna_class:
      mock_dna = Mock()
      mock_dna.has_special_tokens = True
      mock_dna_class.return_value = mock_dna
      
      tokenizer = Tokenizer("dna", 3, special_tokens=None)
      mock_dna_class.assert_called_once_with(kmer=3, continuous=False, special_tokens=None)

  def test_custom_special_tokens(self):
    with patch('biosaic._main.Protein') as mock_protein_class:
      mock_protein = Mock()
      mock_protein.has_special_tokens = True
      mock_protein_class.return_value = mock_protein
      
      custom_tokens = ['<start>', '<end>', '<pad>']
      tokenizer = Tokenizer("protein", 2, special_tokens=custom_tokens)
      mock_protein_class.assert_called_once_with(kmer=2, continuous=False, special_tokens=custom_tokens)

  def test_no_special_tokens(self):
    with patch('biosaic._main.DNA') as mock_dna_class:
      mock_dna = Mock()
      mock_dna.has_special_tokens = False
      mock_dna_class.return_value = mock_dna
      
      tokenizer = Tokenizer("dna", 3, special_tokens=False)
      assert tokenizer.encoding == "dna/base_3k"
      mock_dna_class.assert_called_once_with(kmer=3, continuous=False, special_tokens=False)

class TestEncodingModes:
  def test_continuous_encoding(self):
    with patch('biosaic._main.DNA') as mock_dna_class:
      mock_dna = Mock()
      mock_dna.has_special_tokens = False
      mock_dna_class.return_value = mock_dna
      
      tokenizer = Tokenizer("dna", 4, continuous=True)
      assert tokenizer.encoding == "dna/cont_4k"

  def test_base_encoding(self):
    with patch('biosaic._main.Protein') as mock_protein_class:
      mock_protein = Mock()
      mock_protein.has_special_tokens = False
      mock_protein_class.return_value = mock_protein
      
      tokenizer = Tokenizer("protein", 3, continuous=False)
      assert tokenizer.encoding == "protein/base_3k"

  def test_special_encoding_override(self):
    with patch('biosaic._main.DNA') as mock_dna_class:
      mock_dna = Mock()
      mock_dna.has_special_tokens = True
      mock_dna_class.return_value = mock_dna
      
      tokenizer = Tokenizer("dna", 3, continuous=True)
      assert tokenizer.encoding == "dna/special_3k"

class TestStringRepresentation:
  def test_str_with_special_tokens(self):
    with patch('biosaic._main.Protein') as mock_protein_class:
      mock_protein = Mock()
      mock_protein.has_special_tokens = True
      mock_protein_class.return_value = mock_protein
      
      tokenizer = Tokenizer("protein", 2, continuous=True, special_tokens=['<s>', '</s>'])
      result = str(tokenizer)
      assert "kmer=2" in result
      assert "continuous=True" in result
      assert "special_tokens=2" in result

class TestIntegrationScenarios:
  @patch('biosaic._main.DNA')
  def test_full_dna_workflow(self, mock_dna_class):
    mock_dna = Mock()
    mock_dna.has_special_tokens = False
    mock_dna.encode.return_value = [0, 1, 2]
    mock_dna.decode.return_value = "ATGTGCGCA"
    mock_dna.tokenize.return_value = ["ATG", "TGC", "GCA"]
    mock_dna.vocab_size = 64
    mock_dna_class.return_value = mock_dna
    
    tokenizer = Tokenizer("dna", 3)
    sequence = "ATGTGCGCA"
    
    # Test full encode-decode cycle
    encoded = tokenizer.encode(sequence)
    decoded = tokenizer.decode(encoded)
    tokens = tokenizer.tokenize(sequence)
    
    assert encoded == [0, 1, 2]
    assert decoded == sequence
    assert tokens == ["ATG", "TGC", "GCA"]
    assert tokenizer.vocab_size == 64

  @patch('biosaic._main.Protein')
  def test_protein_with_special_tokens(self, mock_protein_class):
    mock_protein = Mock()
    mock_protein.has_special_tokens = True
    mock_protein.encode.return_value = [1, 5, 10, 2]  # <s>, AA, GG, </s>
    mock_protein.decode.return_value = "<s>AAGG</s>"
    mock_protein_class.return_value = mock_protein
    
    tokenizer = Tokenizer("protein", 2, special_tokens=['<s>', '</s>'])
    sequence = "<s>AAGG</s>"
    
    encoded = tokenizer.encode(sequence)
    decoded = tokenizer.decode(encoded)
    
    assert encoded == [1, 5, 10, 2]
    assert decoded == sequence
    assert tokenizer.encoding == "protein/special_2k"

if __name__ == "__main__":
  pytest.main([__file__, "-v"])


import matplotlib.pyplot as plt
import numpy as np
import random
import math

class CMap:
  def __init__(self, sequence, window_size=1000):
    self.seq = sequence.upper()
    self.window_size = window_size
    self.length = len(sequence)
    self.features = []
    self.orfs = []
    self.repeats = []
    
  def add_feature(self, start, end, name, feature_type='gene', strand='+'):
    colors = {'gene': '#4CAF50', 'cds': '#2196F3', 'rrna': '#FF9800', 'trna': '#9C27B0', 'repeat': '#F44336'}
    self.features.append({'start': start, 'end': end, 'name': name, 'type': feature_type, 'strand': strand, 'color': colors.get(feature_type, '#757575')})
    
  def add_orf(self, start, end, strand='+'): self.orfs.append({'start': start, 'end': end, 'strand': strand})
  def add_repeat(self, start, end, repeat_type='tandem'): self.repeats.append({'start': start, 'end': end, 'type': repeat_type})
  def gc_content(self, seq_chunk): return (seq_chunk.count('G') + seq_chunk.count('C')) / len(seq_chunk) if seq_chunk else 0
  def get_gc_windows(self): return [(self.gc_content(self.seq[i:i+self.window_size]), i) for i in range(0, self.length-self.window_size, self.window_size//8)]
  def pos_to_angle(self, pos): return 2 * np.pi * pos / self.length

  def draw_track_separators(self, ax, radii):
    for r in radii: ax.add_patch(plt.Circle((0, 0), r, fill=False, color='#E0E0E0', linewidth=0.3))

  def draw_feature_labels(self, ax, labels, radius=1.0):
    if not labels: return
    
    # Sort labels by angle
    labels.sort(key=lambda x: x['angle'])
    
    # Resolve overlaps
    min_sep = 0.4  # minimum angular separation
    for _ in range(8):  # max iterations
      overlaps = False
      for i in range(len(labels)):
        for j in range(i+1, len(labels)):
          angle_diff = abs(labels[i]['angle'] - labels[j]['angle'])
          angle_diff = min(angle_diff, 2*np.pi - angle_diff)
          if angle_diff < min_sep:
            overlaps = True
            push = (min_sep - angle_diff) / 2
            labels[i]['angle'] = (labels[i]['angle'] - push) % (2*np.pi)
            labels[j]['angle'] = (labels[j]['angle'] + push) % (2*np.pi)
      if not overlaps: break
    
    # Draw labels
    for label in labels:
      x = label['radius'] * np.cos(label['angle'] - np.pi/2)
      y = label['radius'] * np.sin(label['angle'] - np.pi/2)
      rotation = math.degrees(label['angle']) - 90 if label['angle'] > np.pi else math.degrees(label['angle']) + 90
      if abs(rotation) > 90: rotation += 180
      ax.text(x, y, label['name'], ha='center', va='center', fontsize=6, rotation=rotation, 
              weight='bold', color=label['color'],
              bbox=dict(boxstyle="round,pad=0.5", facecolor="white", alpha=0.8, edgecolor=label['color'], linewidth=0.8))
    major_ticks = max(1, self.length // 12)
    for i in range(0, self.length, major_ticks):
      angle = self.pos_to_angle(i)
      x, y = (radius + 0.08) * np.cos(angle - np.pi/2), (radius + 0.08) * np.sin(angle - np.pi/2)
      label = f"{i//1000000}M" if i >= 1000000 else f"{i//1000}k" if i >= 1000 else str(i)

  def draw_position_labels(self, ax, radius=1.2):
    major_ticks = max(12, self.length // 12)
    for i in range(0, self.length, major_ticks):
      angle = self.pos_to_angle(i)
      x, y = (radius + 0.08) * np.cos(angle - np.pi/2), (radius + 0.08) * np.sin(angle - np.pi/2)
      label = f"{i//1000000}M" if i >= 1000000 else f"{i//1000}k" if i >= 1000 else str(i)
      ax.text(x, y, label, ha='center', va='center', fontsize=6, rotation=math.degrees(angle) - 90 if angle > np.pi else math.degrees(angle) + 90, weight='light')

  def draw_outer_features_track(self, ax, inner_r=2.0, height=0.08):
    labels = []
    for feature in self.features:
      if feature['type'] == 'gene':
        start_angle, end_angle = self.pos_to_angle(feature['start']), self.pos_to_angle(feature['end'])
        radius = inner_r + height/2 + (height/2 if feature['strand'] == '+' else -height/2)
        angles = np.linspace(start_angle, end_angle, max(3, int((end_angle-start_angle)*50)))
        x, y = radius * np.cos(angles - np.pi/2), radius * np.sin(angles - np.pi/2)
        color = '#4169E1' if feature['strand'] == '+' else '#32CD32'
        ax.plot(x, y, color=color, linewidth=1.5, alpha=0.8)
        
        # Add label info
        mid_angle = (start_angle + end_angle) / 2
        labels.append({'angle': mid_angle, 'name': feature['name'], 'color': color, 'radius': inner_r + height + 0.15})
    
    self.draw_feature_labels(ax, labels)

  def draw_inner_features_track(self, ax, inner_r=1.85, height=0.08):
    labels = []
    for feature in self.features:
      if feature['type'] in ['rrna', 'trna']:
        start_angle, end_angle = self.pos_to_angle(feature['start']), self.pos_to_angle(feature['end'])
        radius = inner_r + height/2 + (height/2 if feature['strand'] == '+' else -height/2)
        angles = np.linspace(start_angle, end_angle, max(3, int((end_angle-start_angle)*50)))
        x, y = radius * np.cos(angles - np.pi/2), radius * np.sin(angles - np.pi/2)
        color = '#FF6B35' if feature['type'] == 'rrna' else '#9C27B0'
        ax.plot(x, y, color=color, linewidth=2, alpha=0.9)
        
        # Add label info for rRNA only (tRNA too small/numerous)
        if feature['type'] == 'rrna':
          mid_angle = (start_angle + end_angle) / 2
          labels.append({'angle': mid_angle, 'name': feature['name'], 'color': color, 'radius': inner_r + height + 0.12})
    
    self.draw_feature_labels(ax, labels)

  def draw_gc_skew_track(self, ax, inner_r=1.4, height=0.3):
    gc_data = self.get_gc_windows()
    mean_gc = np.mean([gc for gc, _ in gc_data])
    center_r = inner_r + height/2
    
    angles, values = [], []
    for gc, pos in gc_data:
      angle = self.pos_to_angle(pos)
      skew = (gc - mean_gc) * 2  # amplify for visibility
      angles.append(angle)
      values.append(skew)
    
    # Create continuous line plot
    for i in range(len(angles)-1):
      r1 = center_r + values[i] * height/2
      r2 = center_r + values[i+1] * height/2
      x1, y1 = r1 * np.cos(angles[i] - np.pi/2), r1 * np.sin(angles[i] - np.pi/2)
      x2, y2 = r2 * np.cos(angles[i+1] - np.pi/2), r2 * np.sin(angles[i+1] - np.pi/2)
      color = '#1976D2' if values[i] > 0 else '#D32F2F'
      ax.plot([x1, x2], [y1, y2], color=color, linewidth=0.5, alpha=0.7)

  def draw_gc_content_track(self, ax, inner_r=1.0, height=0.3):
    gc_data = self.get_gc_windows()
    for gc, pos in gc_data:
      angle = self.pos_to_angle(pos)
      gc_height = (gc - 0.3) * height / 0.4
      gc_height = max(0, min(height, gc_height))
      
      if gc > 0.55: color = '#2E7D32'
      elif gc < 0.45: color = '#C62828' 
      else: color = '#1565C0'
        
      x1, y1 = inner_r * np.cos(angle - np.pi/2), inner_r * np.sin(angle - np.pi/2)
      x2, y2 = (inner_r + gc_height) * np.cos(angle - np.pi/2), (inner_r + gc_height) * np.sin(angle - np.pi/2)
      ax.plot([x1, x2], [y1, y2], color=color, linewidth=1, alpha=0.8)

  def draw_repeat_density_track(self, ax, inner_r=0.7, height=0.2):
    density_windows = []
    window_size = self.length // 100
    for i in range(0, self.length, window_size):
      count = sum(1 for r in self.repeats if r['start'] >= i and r['start'] < i + window_size)
      density_windows.append((count, i))
    
    max_density = max([d for d, _ in density_windows]) if density_windows else 1
    for density, pos in density_windows:
      angle = self.pos_to_angle(pos)
      bar_height = (density / max_density) * height
      x1, y1 = inner_r * np.cos(angle - np.pi/2), inner_r * np.sin(angle - np.pi/2)
      x2, y2 = (inner_r + bar_height) * np.cos(angle - np.pi/2), (inner_r + bar_height) * np.sin(angle - np.pi/2)
      ax.plot([x1, x2], [y1, y2], color='#FF5722', linewidth=1.5, alpha=0.6)

  def draw_histogram_tracks(self, ax):
    # Multiple histogram-like tracks with different data
    tracks = [
      {'inner_r': 0.4, 'height': 0.15, 'color': '#E91E63', 'density': 0.3},
      {'inner_r': 0.2, 'height': 0.12, 'color': '#FF9800', 'density': 0.5},
      {'inner_r': 0.05, 'height': 0.1, 'color': '#4CAF50', 'density': 0.7}
    ]
    
    for track in tracks:
      n_bars = int(200 * track['density'])
      for i in range(n_bars):
        angle = 2 * np.pi * i / n_bars + random.uniform(0, 0.1)
        bar_height = random.uniform(0.3, 1.0) * track['height']
        x1, y1 = track['inner_r'] * np.cos(angle - np.pi/2), track['inner_r'] * np.sin(angle - np.pi/2)
        x2, y2 = (track['inner_r'] + bar_height) * np.cos(angle - np.pi/2), (track['inner_r'] + bar_height) * np.sin(angle - np.pi/2)
        ax.plot([x1, x2], [y1, y2], color=track['color'], linewidth=0.8, alpha=0.6)

  def draw_radial_lines(self, ax, inner_r=0.05, outer_r=2.2):
    n_lines = 48
    for i in range(n_lines):
      angle = 2 * np.pi * i / n_lines
      x1, y1 = inner_r * np.cos(angle - np.pi/2), inner_r * np.sin(angle - np.pi/2)
      x2, y2 = outer_r * np.cos(angle - np.pi/2), outer_r * np.sin(angle - np.pi/2)
      ax.plot([x1, x2], [y1, y2], color='#E0E0E0', linewidth=0.2, alpha=0.5)

  def generate_realistic_features(self):
    feature_positions = []
    def is_position_free(start, end, min_gap=200): return all(end < pos_start - min_gap or start > pos_end + min_gap for pos_start, pos_end in feature_positions)

    # Generate genes
    for _ in range(max(15, self.length // 6000)):
      for _ in range(50):  # max attempts
        start = random.randint(0, self.length - 2000)
        length = random.randint(300, 2000)
        end = min(start + length, self.length)
        if is_position_free(start, end):
          self.add_feature(start, end, f"gene_{len(self.features)+1}", 'gene', random.choice(['+', '-']))
          self.add_orf(start, end, random.choice(['+', '-']))
          feature_positions.append((start, end))
          break
    
    # Generate rRNA
    for i, name in enumerate(['16S_rRNA', '23S_rRNA', '5S_rRNA']):
      for _ in range(20):
        start = random.randint(0, self.length - 1200)
        end = min(start + random.randint(800, 1200), self.length)
        if is_position_free(start, end, 500):
          self.add_feature(start, end, name, 'rrna', '+')
          feature_positions.append((start, end))
          break
    
    # Generate tRNA
    aa_list = ['Ala', 'Arg', 'Asn', 'Asp', 'Cys', 'Gln', 'Glu', 'Gly', 'His', 'Ile', 'Leu', 'Lys', 'Met', 'Phe', 'Pro', 'Ser', 'Thr', 'Trp', 'Tyr', 'Val']
    for _ in range(random.randint(20, 30)):
      for _ in range(30):
        start = random.randint(0, self.length - 80)
        end = min(start + random.randint(70, 85), self.length)
        if is_position_free(start, end, 100):
          self.add_feature(start, end, f"tRNA-{random.choice(aa_list)}", 'trna', random.choice(['+', '-']))
          feature_positions.append((start, end))
          break
    
    # Generate repeats
    for _ in range(random.randint(8, 15)):
      start = random.randint(0, self.length - 200)
      end = min(start + random.randint(50, 200), self.length)
      self.add_repeat(start, end)

  def plot(self, figsize=(16, 16), title=None, generate_features=True):
    fig, ax = plt.subplots(figsize=figsize)
    ax.set_xlim(-2.5, 2.5)
    ax.set_ylim(-2.5, 2.5)
    ax.set_aspect('equal')
    ax.axis('off')
    fig.patch.set_facecolor('white')
    
    if generate_features and not self.features: self.generate_realistic_features()
    
    # Draw all tracks from inside out
    track_radii = [0.05, 0.2, 0.4, 0.7, 1.0, 1.4, 1.7, 1.85, 2.0, 2.2]
    self.draw_track_separators(ax, track_radii)
    self.draw_radial_lines(ax)
    self.draw_histogram_tracks(ax)
    self.draw_repeat_density_track(ax)
    self.draw_gc_content_track(ax)
    self.draw_gc_skew_track(ax)
    self.draw_inner_features_track(ax)
    self.draw_outer_features_track(ax)
    self.draw_position_labels(ax)
    
    # Add center circle
    ax.add_patch(plt.Circle((0, 0), 0.05, fill=True, color='white', zorder=10))
    
    plt.tight_layout()
    return fig, ax


import numpy as np
import matplotlib.pyplot as plt

class CGR:
  def __init__(self, size=1000):
    self.size = size
    self.corners = {'A': (0, 0), 'T': (1, 0), 'G': (0, 1), 'C': (1, 1)}
    self.reset()

  def reset(self): self.x, self.y = 0.5, 0.5

  def step(self, base): 
    corner = self.corners.get(base.upper(), (0.5, 0.5))
    self.x, self.y = (self.x + corner[0]) / 2, (self.y + corner[1]) / 2
    return (self.x, self.y)

  def generate_points(self, sequence):
    self.reset()
    return [self.step(base) for base in sequence if base.upper() in self.corners]

  def create_heatmap(self, points):
    grid = np.zeros((self.size, self.size))
    coords = [(int(y * (self.size - 1)), int(x * (self.size - 1))) for x, y in points]
    for i, j in coords:
      if 0 <= i < self.size and 0 <= j < self.size: grid[i, j] += 1
    return grid

  def create_multicolor_heatmap(self, sequence):
    # create separate point lists for each base
    base_points = {'A': [], 'T': [], 'G': [], 'C': []}
    self.reset()
    
    for base in sequence:
      if base.upper() in self.corners:
        x, y = self.step(base)
        base_points[base.upper()].append((x, y))
    
    # create rgb image
    rgb_image = np.zeros((self.size, self.size, 3))

    # map each base to color channels with proper intensity
    for base, color_channel in [('A', 0), ('T', 1), ('G', 2)]:  # red, green, blue
      if base_points[base]:
        coords = [(int(y * (self.size - 1)), int(x * (self.size - 1))) for x, y in base_points[base]]
        for i, j in coords:
          if 0 <= i < self.size and 0 <= j < self.size:
            rgb_image[i, j, color_channel] += 1

    # handle C (cytosine) as magenta (red + blue)
    if base_points['C']:
      coords = [(int(y * (self.size - 1)), int(x * (self.size - 1))) for x, y in base_points['C']]
      for i, j in coords:
        if 0 <= i < self.size and 0 <= j < self.size:
          rgb_image[i, j, 0] += 0.8  # red component
          rgb_image[i, j, 2] += 0.8  # blue component

    # normalize each channel independently
    for c in range(3):
      if rgb_image[:, :, c].max() > 0:
        rgb_image[:, :, c] = rgb_image[:, :, c] / rgb_image[:, :, c].max()

    # apply gamma correction and brightness boost
    rgb_image = np.power(rgb_image, 0.8) * 1.5
    return np.clip(rgb_image, 0, 1)

  def visualize(self, sequence, title="CGR Visualization", cmap='hot', figsize=(12, 10), label=False, multicolor=True):
    plt.figure(figsize=figsize)

    if multicolor:
      rgb_heatmap = self.create_multicolor_heatmap(sequence)
      plt.imshow(rgb_heatmap, origin='lower', extent=[0, 1, 0, 1])
      plt.title(f"{title}\nSequence length: {len(sequence)} (A=Red, T=Green, G=Blue, C=Magenta)")
    else:
      points = self.generate_points(sequence)
      heatmap = self.create_heatmap(points)
      # apply log scaling for better visualization
      heatmap_log = np.log1p(heatmap)  # log(1+x) to handle zeros
      plt.imshow(heatmap_log, cmap=cmap, origin='lower', extent=[0, 1, 0, 1])
      plt.colorbar(label='log(frequency + 1)')
      plt.title(f"{title}\nSequence length: {len(sequence)}")

    plt.xlabel('X'); plt.ylabel('Y')

    # add corner labels with better visibility
    if label:
      corner_props = [('A', 0.02, 0.02, 'red'), ('T', 0.98, 0.02, 'lime'), ('G', 0.02, 0.98, 'cyan'), ('C', 0.98, 0.98, 'magenta')]
      for text, x, y, color in corner_props:
        plt.text(x, y, text, fontsize=16, fontweight='bold', color=color, bbox=dict(boxstyle='circle,pad=0.3', facecolor='black', alpha=0.8, edgecolor=color))

    plt.tight_layout()
    return plt


from itertools import product
import json, pickle
import os, tempfile, urllib, requests
import numpy as np
import re

AMINO_ACIDS = [
  'A','R','N','D','C','Q','E','G','H','I',
  'L','K','M','F','P','S','T','W','Y','V','-'
]

class Protein:
  def __init__(self, kmer: int, continuous: bool=True, special_tokens=None):
    self.kmer = kmer
    self.continuous = continuous
    self._base_chars = AMINO_ACIDS
    self._ids_to_taken, self.vocab = {}, {}

    # handle special tokens
    self.init_special_tokens = ['<S>', '</S>', '<P>', '<C>', '<M>']

    # setting has_special_tokens first based on continuous mode
    self.has_special_tokens = not self.continuous
    if special_tokens is False:
      self.special_tokens = []
      self.has_special_tokens = False
    elif special_tokens is None:
      # using default special tokens only if continuous=False
      self.special_tokens = self.init_special_tokens if not self.continuous else []
    else:
      self.special_tokens = special_tokens

    # special tokens only work with continuous=False
    if self.special_tokens and continuous:
      raise ValueError("Special tokens are only supported with continuous=False")

    if self.continuous:
      self.vocab_size = len(self._base_chars) ** kmer
    else:
      self.vocab_size = sum(len(self._base_chars) ** i for i in range(1, kmer+1))
    if self.has_special_tokens:
      self.vocab_size += len(self.special_tokens)

  def _split_with_special_tokens(self, sequence):
    """Split sequence preserving special tokens"""
    if not self.has_special_tokens:
      return [sequence]

    # Create regex pattern for all special tokens
    pattern = '(' + '|'.join(re.escape(token) for token in self.special_tokens) + ')'
    parts = re.split(pattern, sequence)
    return [part for part in parts if part]  # Remove empty strings

  def tokenize(self, sequence):
    if not self.has_special_tokens:
      if any(ch not in self._base_chars for ch in sequence):
        raise ValueError("Invalid character in Protein sequence")
      return [sequence[i:i+self.kmer] for i in range(len(sequence) - self.kmer + 1)] if self.continuous else [sequence[i:i+self.kmer] for i in range(0, len(sequence), self.kmer)]

    # Handle special tokens (only works with continuous=False)
    tokens = []
    for part in self._split_with_special_tokens(sequence):
      if part in self.special_tokens:
        tokens.append(part)
      else:
        if any(ch not in self._base_chars for ch in part):
          raise ValueError("Invalid character in Protein sequence")
        tokens.extend([part[i:i+min(self.kmer, len(part)-i)] for i in range(0, len(part), self.kmer) if i < len(part)])
    return tokens

  def detokenize(self, ids):
    if self.continuous and not self.has_special_tokens:
      return "" if not ids else "".join(ids[i][0] for i in range(len(ids))) + ids[-1][1:]
    return "".join(ids)

  def build_vocab(self):
    letters, combos = sorted(self._base_chars), []
    if self.continuous:
      combos = list(product(letters, repeat=self.kmer))
    else:
      for L in range(1, self.kmer + 1):
        combos.extend(product(letters, repeat=L))

    self.vocab = {''.join(c): i for i, c in enumerate(combos)}
    if self.has_special_tokens:
      start_idx = len(self.vocab)
      for i, token in enumerate(self.special_tokens):
        self.vocab[token] = start_idx + i
    self.ids_to_token = {v: k for k, v in self.vocab.items()}
    self.vocab_size = len(self.vocab)

  def encode(self, sequence):
    sequence = sequence.upper()
    tokenized_data = self.tokenize(sequence)
    return [self.vocab[kmer] for kmer in tokenized_data if kmer in self.vocab]

  def decode(self, ids):
    tokens = self.ids_to_chars(ids)
    return self.detokenize(tokens)

  def ids_to_chars(self, ids: list[int]):
    assert isinstance(ids, list) and len(ids) > 0, "ids must be a non-empty list"
    assert isinstance(ids[0], int), "only accepts encoded ids"
    return [self.ids_to_token[i] for i in ids]

  def chars_to_ids(self, chars: list[str]):
    assert isinstance(chars, list) and len(chars) > 0, "chars must be a non-empty list"
    assert isinstance(chars[0], str), "only accepts tokenized strings"
    return [self.vocab[i] for i in chars]

  def verify(self, ids, file=None):
    verified = []
    ids = self.ids_to_chars(ids) if isinstance(ids[0], int) else ids
    for i in range(len(ids) - 1):
      # Skip verification for special tokens
      if ids[i] in self.special_tokens or ids[i+1] in self.special_tokens:
        verified.append({"kmer1": ids[i], "kmer2": ids[i + 1], "match": "special_token"})
        continue
      match = ids[i][1:] == ids[i + 1][:-1] if self.continuous else True
      verified.append({"kmer1": ids[i], "kmer2": ids[i + 1], "match": match})
    
    if file:
      with open(os.path.join(file, "verify.json"), 'w', encoding='utf-8') as f:
        json.dump(verified, f)
    return verified

  def save(self, path, as_json=False):
    os.makedirs(os.path.dirname(path), exist_ok=True)
    data = {
      "kmer": self.kmer,
      "vocab_size": self.vocab_size,
      "trained_vocab": self.vocab,
      "special_tokens": self.special_tokens,
      "continuous": self.continuous
    }
    ext = ".json" if as_json else ".model"
    with open(path + ext, "w" if as_json else "wb", encoding="utf-8" if as_json else None) as f:
      json.dump(data, f, indent=2) if as_json else pickle.dump(data, f)
    print(f"DEBUGG INFO[104] [Saved] Vocabulary saved to {path + ext}")

  def load(self, model_path: str):
    def is_url(path):
      return path.startswith(("http://", "https://"))

    if is_url(model_path):
      with tempfile.NamedTemporaryFile(delete=False, suffix=".model" if model_path.endswith(".model") else ".json") as tmp_file:
        try:
          urllib.request.urlretrieve(model_path.replace("blob/", ""), tmp_file.name)
          model_path = tmp_file.name
        except Exception as e:
          raise RuntimeError(f"Failed to download model from {model_path}: {e}")

    with open(model_path, "r" if model_path.endswith(".json") else "rb", encoding="utf-8" if model_path.endswith(".json") else None) as f:
      data = json.load(f) if model_path.endswith(".json") else pickle.load(f)

    if not model_path.endswith((".json", ".model")):
      raise TypeError("Only supports vocab file format `.model` & `.json`")

    self.vocab = data["trained_vocab"]
    self.kmer = data.get("kmer", self.kmer)
    self.continuous = data.get("continuous", self.continuous)
    self.ids_to_token = {v: k for k, v in self.vocab.items()}
    self.has_special_tokens = not self.continuous
    if self.has_special_tokens:
      loaded_special_tokens = self.special_tokens
      self.special_tokens = list(dict.fromkeys(data.get("special_tokens", []) + loaded_special_tokens))

      max_id = max(self.vocab.values(), default=-1)
      for token in self.special_tokens:
        if token not in self.vocab:
          max_id += 1
          self.vocab[token] = max_id
          self.ids_to_token[max_id] = token
    self.vocab_size = len(self.vocab)

  def one_hot_encode(self, sequence):
    tokens = self.tokenize(sequence.upper())
    print(len(tokens), tokens)
    one_hot = np.zeros((len(tokens), len(self.vocab)), dtype=int)
    for i, token in enumerate(tokens):
      if token in self.vocab:
        one_hot[i, self.vocab[token]] = 1
    return one_hot

  def pad_sequence(self, sequence, target_length, pad_char='-'):
    return sequence[:target_length] if len(sequence) >= target_length else sequence + pad_char * (target_length - len(sequence))

  def reverse_complement(self, sequence):
    raise NotImplementedError("Proteins don't have reverse complement! You dumbass!!!")


from ._dna import DNA
from ._protein import Protein
from ._rna import RNA
from typing import List

main_base_url = "https://raw.githubusercontent.com/delveopers/biosaic/main/vocab/"
dev_base_url = "https://raw.githubusercontent.com/delveopers/biosaic/dev/vocab/"
hugginface_url = "https://huggingface.co/shivendrra/BiosaicTokenizer/resolve/main/kmers/"

class Tokenizer:
  """
    Biosaic Tokenizer class for DNA/RNA and Protein sequences with special token support.

    This class wraps around DNA/RNA and Protein tokenizers, allowing encoding,
    decoding, tokenization, and detokenization of biological sequences 
    using pre-trained vocabularies stored remotely, with optional special tokens.

    Attributes:
      kmer (int): The k-mer size used for tokenization.
      continuous (bool): Whether to use a sliding-window tokenization (`True`) or fixed-length non-overlapping (`False`).
      special_tokens (list or None): List of special tokens to include in vocabulary.
      encoding (str): The encoding identifier used to locate and load vocab files.
      encoding_path (str): URL path pointing to the pretrained vocabulary model file.
      _tokenizer (DNA/RNA or Protein): Internal tokenizer instance specific to the sequence type.
  """
  def __init__(self, mode: str, kmer: int, continuous: bool=False, special_tokens=None):
    """
      Initializes the Tokenizer with the specified mode, k-mer length, tokenization style, and special tokens.

      Args:
        mode (str): Type of sequence to tokenize. Should be either "dna", "rna" or "protein".
        kmer (int): The k-mer length used for tokenization. Maximum allowed is 8 for DNA/RNA and 4 for protein.
        continuous (bool): If True, enables sliding-window tokenization (i.e., overlapping k-mers).
                          If False, tokenizes in fixed non-overlapping k-mer chunks.
        special_tokens (list or None): List of special tokens to add to vocabulary. 
                                      If None, uses default ['<s>', '</s>', '<p>', '<c>', '<m>'].
                                      If False, no special tokens are added.

      Raises:
        AssertionError: If an invalid mode is specified or k-mer size is above supported limit.
        ValueError: If special tokens are used with continuous=True.
    """
    assert (mode == "dna" or mode == "rna" or mode == "protein"), "Unknown mode type, choose b/w ``dna``, ``rna`` or ``protein``"
    if mode == "protein": assert (kmer <= 4), "KMer size supported only till 4 for protein!"
    else: assert (kmer <= 8), "KMer size supported only till 8 for DNA or RNA!"
    self.kmer, self.continuous, self.special_tokens = kmer, continuous, special_tokens

    if mode == "dna": self._tokenizer = DNA(kmer=kmer, continuous=continuous, special_tokens=special_tokens)
    elif mode == "rna": self._tokenizer = RNA(kmer=kmer, continuous=continuous, special_tokens=special_tokens)
    else: self._tokenizer = Protein(kmer=kmer, continuous=continuous, special_tokens=special_tokens)

    if continuous: self.encoding = f"{mode}/cont_{kmer}k"
    else: self.encoding = f"{mode}/base_{kmer}k"

    self.encoding_path = main_base_url + self.encoding + ".model"
    self._tokenizer.load(model_path=self.encoding_path)
    self.encoding = f"{mode}/special_{kmer}k" if self._tokenizer.has_special_tokens else self.encoding

  def encode(self, sequence: str) -> list[int]:
    """
      Encodes a biological sequence into integer token IDs.

      Args:
        sequence (str): DNA/RNA or protein sequence composed of valid characters and/or special tokens.
      Returns:
        List[int]: Encoded token IDs corresponding to k-mers and special tokens in the sequence.
      Raises:
        ValueError: If the input sequence contains invalid characters.
    """
    return self._tokenizer.encode(sequence)
  
  def decode(self, ids: list[int]) -> str:
    """
      Decodes a list of token IDs back into the original sequence.

      Args:
        ids (List[int]): Encoded token IDs representing a biological sequence.
      Returns:
        str: Decoded DNA/protein sequence reconstructed from k-mers and special tokens.
    """
    return self._tokenizer.decode(ids)

  def tokenize(self, sequence: str) -> List[str]:
    """
      Splits the input biological sequence into k-mer tokens and special tokens.

      Args:
        sequence (str): DNA/RNA or protein string potentially containing special tokens.
      Returns:
        List[str]: List of k-mer substrings and special tokens, with variable lengths when special tokens present.
    """
    return self._tokenizer.tokenize(sequence)

  def detokenize(self, ids: List[str]) -> str:
    """
      Combines k-mer tokens and special tokens into the original sequence.

      Args:
        ids (List[str]): List of k-mer tokens and special tokens.
      Returns:
        str: Reconstructed sequence from tokenized substrings.
    """
    return self._tokenizer.detokenize(ids)

  def one_hot(self, sequence):
    return self._tokenizer.one_hot_encode(sequence)

  def reverse_complement(self, sequence):
    return self._tokenizer.reverse_complement(sequence)

  def pad_sequence(self, sequence, target_length, pad_char="-"):
    return self._tokenizer.pad_sequence(sequence, target_length, pad_char)

  @property
  def vocab_size(self):
    return self._tokenizer.vocab_size

  @property
  def vocab(self):
    return self._tokenizer.vocab

  def __str__(self):
    special_info = f", special_tokens={len(self.special_tokens) if self.special_tokens else 0}" if hasattr(self, 'special_tokens') else ""
    return f"biosaic.tokenizer <kmer={self.kmer}, encoding={self.encoding}, continuous={self.continuous}{special_info}>"


import pytest
import tempfile
import os
import json
import numpy as np
from biosaic._protein import Protein, AMINO_ACIDS

class TestProteinConstants:
  def test_amino_acids_constant(self):
    expected = ['A','R','N','D','C','Q','E','G','H','I','L','K','M','F','P','S','T','W','Y','V','-']
    assert AMINO_ACIDS == expected
    assert len(AMINO_ACIDS) == 21

class TestProteinInitialization:
  def test_init_continuous(self):
    protein = Protein(kmer=3, continuous=True)
    assert protein.kmer == 3
    assert protein.continuous == True
    assert protein.vocab_size == 21**3
    assert protein.has_special_tokens == False

  def test_init_non_continuous(self):
    protein = Protein(kmer=3, continuous=False)
    assert protein.kmer == 3
    assert protein.continuous == False
    expected_size = 21 + 441 + 9261 + len(protein.special_tokens)
    assert protein.vocab_size == expected_size

  def test_init_with_special_tokens(self):
    special_tokens = ['<START>', '<END>']
    protein = Protein(kmer=2, continuous=False, special_tokens=special_tokens)
    assert protein.has_special_tokens == True
    assert protein.special_tokens == special_tokens
    expected_size = (21 + 441) + 2
    assert protein.vocab_size == expected_size

  def test_special_tokens_with_continuous_raises_error(self):
    with pytest.raises(ValueError):
      Protein(kmer=3, continuous=True, special_tokens=['<START>'])

  def test_large_kmer_vocab_size_calculation(self):
    protein = Protein(kmer=4, continuous=True)
    assert protein.vocab_size == 21**4
    
    protein_nc = Protein(kmer=4, continuous=False)
    expected_size = 21 + 21**2 + 21**3 + 21**4 + len(protein_nc.special_tokens)
    assert protein_nc.vocab_size == expected_size

class TestProteinTokenization:
  def test_tokenize_continuous(self):
    protein = Protein(kmer=3, continuous=True)
    tokens = protein.tokenize("ARNDCQ")
    assert tokens == ['ARN', 'RND', 'NDC', 'DCQ']

  def test_tokenize_non_continuous(self):
    protein = Protein(kmer=3, continuous=False)
    tokens = protein.tokenize("ARNDCQE")
    assert tokens == ['ARN', 'DCQ', 'E']

  def test_tokenize_with_special_tokens(self):
    protein = Protein(kmer=2, continuous=False, special_tokens=['<S>', '</S>'])
    tokens = protein.tokenize("<S>ARND</S>")
    assert tokens == ['<S>', 'AR', 'ND', '</S>']

  def test_tokenize_mixed_special_tokens(self):
    protein = Protein(kmer=2, continuous=False, special_tokens=['<P>', '<M>'])
    tokens = protein.tokenize("AR<P>ND<M>CQ")
    assert tokens == ['AR', '<P>', 'ND', '<M>', 'CQ']

  def test_tokenize_invalid_character(self):
    protein = Protein(kmer=2)
    with pytest.raises(ValueError):
      protein.tokenize("ARNDX")

  def test_all_amino_acids_tokenization(self):
    protein = Protein(kmer=1, continuous=False)
    protein.build_vocab()
    sequence = "ARNDCQEGHILKMFPSTWYV-"
    tokens = protein.tokenize(sequence)
    assert len(tokens) == 21
    assert tokens == list(sequence)

  def test_case_insensitive_processing(self):
    protein = Protein(kmer=2, continuous=False)
    protein.build_vocab()
    encoded_lower = protein.encode("arnd")
    encoded_upper = protein.encode("ARND")
    assert encoded_lower == encoded_upper

class TestProteinVocabulary:
  def test_build_vocab_continuous(self):
    protein = Protein(kmer=2, continuous=True)
    protein.build_vocab()
    assert len(protein.vocab) == 441
    assert 'AA' in protein.vocab
    assert 'RR' in protein.vocab
    assert '--' in protein.vocab

  def test_build_vocab_non_continuous(self):
    protein = Protein(kmer=2, continuous=False)
    protein.build_vocab()
    assert len(protein.vocab) == 467
    assert 'A' in protein.vocab
    assert 'AA' in protein.vocab

  def test_build_vocab_with_special_tokens(self):
    protein = Protein(kmer=2, continuous=False, special_tokens=['<S>', '</S>'])
    protein.build_vocab()
    assert len(protein.vocab) == 464  # 462 + 2
    assert '<S>' in protein.vocab
    assert '</S>' in protein.vocab

class TestProteinEncodeDecode:
  def test_encode_decode_roundtrip(self):
    protein = Protein(kmer=2, continuous=True)
    protein.build_vocab()
    sequence = "ARNDCQEGHIL"
    encoded = protein.encode(sequence)
    decoded = protein.decode(encoded)
    assert decoded == sequence

  def test_encode_decode_with_special_tokens(self):
    protein = Protein(kmer=2, continuous=False, special_tokens=['<S>', '</S>'])
    protein.build_vocab()
    sequence = "<S>ARND</S>"
    encoded = protein.encode(sequence)
    decoded = protein.decode(encoded)
    assert decoded == sequence

  def test_ids_to_chars_and_chars_to_ids(self):
    protein = Protein(kmer=2, continuous=True)
    protein.build_vocab()
    chars = ['AR', 'RN', 'ND']
    ids = protein.chars_to_ids(chars)
    converted_chars = protein.ids_to_chars(ids)
    assert chars == converted_chars

class TestProteinDetokenization:
  def test_detokenize_continuous(self):
    protein = Protein(kmer=3, continuous=True)
    tokens = ['ARN', 'RND', 'NDC']
    result = protein.detokenize(tokens)
    assert result == "ARNDC"
    
    assert protein.detokenize(['ARN']) == "ARN"
    assert protein.detokenize([]) == ""

  def test_detokenize_non_continuous(self):
    protein = Protein(kmer=2, continuous=False)
    tokens = ['AR', 'ND', 'CQ']
    result = protein.detokenize(tokens)
    assert result == "ARNDCQ"

class TestProteinVerification:
  def test_verify_continuous_valid(self):
    protein = Protein(kmer=3, continuous=True)
    protein.build_vocab()
    tokens = ['ARN', 'RND', 'NDC']
    verified = protein.verify(tokens)
    assert verified[0]['match'] == True
    assert verified[1]['match'] == True

  def test_verify_continuous_invalid(self):
    protein = Protein(kmer=3, continuous=True)
    protein.build_vocab()
    tokens = ['ARN', 'CQE']
    verified = protein.verify(tokens)
    assert verified[0]['match'] == False

  def test_verify_with_special_tokens(self):
    protein = Protein(kmer=2, continuous=False, special_tokens=['<S>'])
    protein.build_vocab()
    tokens = ['<S>', 'AR', 'ND']
    verified = protein.verify(tokens)
    assert verified[0]['match'] == 'special_token'
    assert verified[1]['match'] == True

class TestProteinPersistence:
  @pytest.fixture
  def temp_dir(self):
    temp_dir = tempfile.mkdtemp()
    yield temp_dir
    for file in os.listdir(temp_dir):
      os.remove(os.path.join(temp_dir, file))
    os.rmdir(temp_dir)

  def test_save_and_load_json(self, temp_dir):
    protein = Protein(kmer=2, continuous=True)
    protein.build_vocab()
    save_path = os.path.join(temp_dir, "test_protein_vocab")
    protein.save(save_path, as_json=True)
    
    protein2 = Protein(kmer=2, continuous=True)
    protein2.load(save_path + ".json")
    
    assert protein.kmer == protein2.kmer
    assert protein.continuous == protein2.continuous
    assert protein.vocab == protein2.vocab

  def test_save_and_load_pickle(self, temp_dir):
    protein = Protein(kmer=2, continuous=False, special_tokens=['<TEST>'])
    protein.build_vocab()
    save_path = os.path.join(temp_dir, "test_protein_vocab")
    protein.save(save_path, as_json=False)
    
    protein2 = Protein(kmer=2)
    protein2.load(save_path + ".model")
    
    assert protein.kmer == protein2.kmer
    assert protein.continuous == protein2.continuous
    assert protein.special_tokens == protein2.special_tokens
    assert protein.vocab == protein2.vocab

  def test_verify_save_to_file(self, temp_dir):
    protein = Protein(kmer=2, continuous=True)
    protein.build_vocab()
    tokens = ['AR', 'RN']
    verified = protein.verify(tokens, file=temp_dir)
    
    verify_file = os.path.join(temp_dir, "verify.json")
    assert os.path.exists(verify_file)
    
    with open(verify_file, 'r') as f:
      saved_data = json.load(f)
    assert len(saved_data) == 1
    assert saved_data[0]['kmer1'] == 'AR'
    assert saved_data[0]['kmer2'] == 'RN'

class TestProteinUtilities:
  def test_one_hot_encode(self):
    protein = Protein(kmer=2, continuous=False)
    protein.build_vocab()
    one_hot = protein.one_hot_encode("ARND")
    
    expected_shape = (2, len(protein.vocab))
    assert one_hot.shape == expected_shape
    for row in one_hot:
      assert np.sum(row) == 1

  def test_pad_sequence(self):
    protein = Protein(kmer=2)
    assert protein.pad_sequence("ARN", 6) == "ARN---"
    assert protein.pad_sequence("ARNDCQEGHIL", 5) == "ARNDC"
    assert protein.pad_sequence("ARN", 3) == "ARN"
    assert protein.pad_sequence("AR", 5, pad_char='X') == "ARXXX"

  def test_reverse_complement_not_implemented(self):
    protein = Protein(kmer=2)
    with pytest.raises(NotImplementedError, match="Proteins don't have reverse complement"):
      protein.reverse_complement("ARND")

class TestProteinEdgeCases:
  def test_edge_cases(self):
    protein = Protein(kmer=1, continuous=True)
    protein.build_vocab()
    tokens = protein.tokenize("ARND")
    assert tokens == ['A', 'R', 'N', 'D']
    
    protein = Protein(kmer=2, continuous=False)
    assert protein.tokenize("") == []
    assert protein.tokenize("A") == ['A']

if __name__ == '__main__':
  pytest.main([__file__, "-v"])